#include "license_reader.h"
#include "license_logging.h"
#include "license_error.h"
#include "license_backbox.h"
#include "backbox/ezxml.h"
#include "xalloc.h"
#include <nss.h>
#include <cms.h>
#include <cert.h>
#include <certdb.h>
#include <base64.h>
#include <prerror.h>
#include <lauxlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>

#define LICENSE_BLOCK_HEADER "-----BEGIN OP5 LICENSE BLOCK-----"
#define LICENSE_BLOCK_FOOTER "-----END OP5 LICENSE BLOCK-----"

#include "trusted.c"

struct op5lic_reader_t {
	char *lua_buffer;
	char *filename;
	FILE *fp;
	NSSCMSMessage *cms_message;
	CERTCertificate **trusted_certificates;
	int ntrusted;
	LicenseStatus status;
};

struct op5lic_license_buffer {
	char * buffer; /*(base64-encoded) CMSSignedData*/
	size_t len;
};

static struct certificate_options_t {
	NSSInitContext *nss_context;
	CERTCertDBHandle *cert_db_handle;
} options;

static int nss_init(void) {

	/* flags set to reflect the options used by NSS_NoDB_Init */
	int nss_flags = ( NSS_INIT_READONLY
			| NSS_INIT_NOCERTDB | NSS_INIT_NOMODDB
			| NSS_INIT_FORCEOPEN | NSS_INIT_OPTIMIZESPACE
			);

	NSSInitContext *nss_context = NSS_InitContext(
			"", /* configdir */
			"", /* cerPrefix */
			"", /* keyPrefix */
			"", /* secmodName */
			NULL, /*initParams*/
			nss_flags

			);

	if (!nss_context) {
		log_err("NSS Init Context failed: %s", PR_ErrorToName(PR_GetError()));
		return -1;
	}

	if (CERT_GetUsePKIXForValidation() == PR_TRUE) {
		/* Just in case the default changes... */
		log_err("Using PKIX for validation is not supported");
		return -1;
	}
	options.nss_context = nss_context;
	options.cert_db_handle = CERT_GetDefaultCertDB();
	return 0;
}

static void nss_deinit(void) {
	if ( SECSuccess != NSS_ShutdownContext(options.nss_context)) {
		log_err("NSS Shutdown failed %s", PR_ErrorToName(PR_GetError()));
	}
}

struct op5lic_license_buffer * op5lic_license_buffer_create(void) {
	struct op5lic_license_buffer *new_lb;
	new_lb = xcalloc(1, sizeof(*new_lb));
	new_lb->buffer = NULL;
	new_lb->len = 0;
	return new_lb;
}

void op5lic_license_buffer_destroy(struct op5lic_license_buffer *lb) {
	if (!lb)
		return;

	xfree(lb->buffer);
	xfree(lb);
}

struct op5lic_license_buffer * op5lic_license_buffer_from_file(FILE *fp) {
	char *buf = NULL;
	struct op5lic_license_buffer * lb = op5lic_license_buffer_create();
	const char *header = NULL;
	const char *footer = NULL;
	struct stat statp;
	memset(&statp, 0, sizeof(statp));

	char *dest = NULL;
	const char *p = NULL;

	if (!fp)
		return NULL;

	if (fstat(fileno(fp), &statp)) {
		log_err("Failed to stat FP: %s", strerror(errno));
		op5lic_license_buffer_destroy(lb);
		return NULL;
	}
	if (statp.st_size < (unsigned int)(strlen(LICENSE_BLOCK_HEADER) + strlen(LICENSE_BLOCK_FOOTER))) {
		log_err("License file impossibly small, invalid license");
		op5lic_license_buffer_destroy(lb);
		return NULL;
	}

	buf = xcalloc(1, statp.st_size + 1 );
	if (fread(buf, 1, statp.st_size, fp) != (unsigned int) statp.st_size) {
		log_err("Could not read entire license buffer");
		op5lic_license_buffer_destroy(lb);
		xfree(buf);
		return NULL;
	}
	buf[statp.st_size] = '\0';
	if ((header = strstr(buf, LICENSE_BLOCK_HEADER)) == NULL) {
		log_warn("Could not find license block header, invalid license");
		op5lic_license_buffer_destroy(lb);
		xfree(buf);
		return NULL;
	}

	p = header + strlen(LICENSE_BLOCK_HEADER);
	if ((footer = strstr(p, LICENSE_BLOCK_FOOTER)) == NULL) {
		log_warn("Could not find license block footer, invalid license");
		op5lic_license_buffer_destroy(lb);
		xfree(buf);
		return NULL;
	}

	dest = xcalloc(1, (footer - header) + 1);
	lb->buffer = dest;
	while (*p != '\0' && p < footer) {
		switch (*p) {
			case '\n':
			case '\r':
				/*ignore*/
				break;
			default:
				*dest = *p;
				dest++;
		}
		++p;
	}
	*dest = '\0';
	xfree(buf);
	return lb;
}

struct op5lic_reader_t *op5lic_reader_create(const char *filename, struct op5lic_error_t *err) {
	struct op5lic_reader_t *lr;

	lr = xcalloc(1, sizeof(struct op5lic_reader_t));
	if (nss_init() < 0) {
		log_err("NSS initialization failure");
		op5lic_error_set(err, OP5LIC_STATUS_NSS_FAILURE, "NSS initialization failure");
		goto error;
	}

	if (options.cert_db_handle == NULL) {
		log_err("NSS initialization failure: no cert DB set");
		op5lic_error_set(err, OP5LIC_STATUS_NSS_FAILURE, "NSS initialization failure: no cert DB set");
		goto error;
	}

	if (options.nss_context == NULL) {
		log_err("NSS initialization failure: no NSS context set");
		op5lic_error_set(err, OP5LIC_STATUS_NSS_FAILURE, "NSS initialization failure: no NSS context set");
		goto error;
	}

	log_info("Opening license file: %s", filename);
	lr->filename = xstrdup(filename);
	lr->fp = fopen(lr->filename, "r");
	if(!lr->fp) {
		op5lic_error_set(err, OP5LIC_STATUS_NOT_EXIST, filename);
		log_err("Failed to open license file \"%s\"", filename);
		goto error;
	}
	int i;
	for (i = 0; trusted_certs[i] != NULL; i++) {
		CERTCertificate *cert = CERT_ConvertAndDecodeCertificate(trusted_certs[i]); /* Also creates TempCert*/
		if (cert == NULL) {
			log_warn("Failed to decode trusted certificate: %s", trusted_certs[i]);
			continue;
		}
		/* We trust this cert to authorize email signing*/
		CERTCertTrust trust;
		trust.sslFlags = 0;
		trust.emailFlags = CERTDB_TRUSTED_CA | CERTDB_VALID_CA;
		trust.objectSigningFlags = 0;

		if (CERT_ChangeCertTrust(options.cert_db_handle, cert, &trust)) {
			log_err("Failed to modify trust attribute of CA %s: %s", cert->subjectName, PR_ErrorToName(PR_GetError()));
			continue;
		}
		lr->trusted_certificates = xrealloc(lr->trusted_certificates, (lr->ntrusted+1) * sizeof(cert));
		lr->trusted_certificates[lr->ntrusted++] = cert;
	}
	return lr;

error:
	op5lic_reader_destroy(lr);
	return NULL;
}

void op5lic_reader_destroy(struct op5lic_reader_t *lr) {
	if (lr == NULL)
		return;

	xfree(lr->filename);
	xfree(lr->lua_buffer);

	if (lr->fp) {
		fclose(lr->fp);
		lr->fp = NULL;
	}
	if (lr->cms_message) {
		NSS_CMSMessage_Destroy(lr->cms_message);
		lr->cms_message = NULL;
	}

	if (lr->ntrusted > 0) {
		int i;
		for (i = 0; i < lr->ntrusted; i++) {
			CERT_DestroyCertificate(lr->trusted_certificates[i]);
		}
		xfree(lr->trusted_certificates);
	}
	xfree(lr);
	nss_deinit();
}


static NSSCMSMessage *
op5lic_reader_decode_license(struct op5lic_reader_t *lr) {
	NSSCMSDecoderContext *p7dcx = NULL;
	struct op5lic_license_buffer *licbuf= NULL;
	unsigned char *decoded_buffer = NULL;
	unsigned int decoded_buffer_len = 0;
	SECStatus rv;
	if (lr->cms_message) {
		/* short circuit, we've already decoded the message */
		return lr->cms_message;
	}

	if (!lr || !(lr->fp)) {
		log_err("Missing reader or license FP");
		return NULL;
	}

	if (!options.cert_db_handle) {
		log_err("Missing cert DB");
		return NULL;
	}

	licbuf = op5lic_license_buffer_from_file(lr->fp);
	if (!licbuf) {
		log_err("Failed to read license");
		return NULL;
	}

	decoded_buffer = ATOB_AsciiToData(licbuf->buffer, &decoded_buffer_len);
	if (decoded_buffer == NULL) {
		log_err("Failed to base64-decode license buffer");
		op5lic_license_buffer_destroy(licbuf);
		return NULL;
	}

	p7dcx = NSS_CMSDecoder_Start(
			NULL, /*area*/
			NULL, NULL, /*content CB*/
			NULL, NULL, /*password CB*/
			NULL, NULL); /*decrypt key CB*/

	if(!p7dcx) {
		log_err("Failed to start decoder");
		op5lic_license_buffer_destroy(licbuf);
		return NULL;
	}

	rv = NSS_CMSDecoder_Update(p7dcx, (char *)decoded_buffer, decoded_buffer_len);
	PORT_Free(decoded_buffer);
	op5lic_license_buffer_destroy(licbuf);

	if ( rv != SECSuccess ) {
		log_err("Failed to update decoder");
		NSS_CMSDecoder_Cancel(p7dcx);
		return NULL;
	}

	lr->cms_message = NSS_CMSDecoder_Finish(p7dcx);
	return lr->cms_message;
}

static bool
op5lic_reader_license_is_signed(NSSCMSMessage *cms_message) {
	return NSS_CMSMessage_IsSigned(cms_message);
}

static bool
op5lic_reader_license_signature_is_valid(NSSCMSMessage *cms_message, CERTCertificate **trusted_certificates, int ntrusted) {
	if (!cms_message)
		return false;

	if (!trusted_certificates || ntrusted == 0) {
		log_err("No trusted certificate supplied, signature will never be valid");
		return false;
	}
	return true;
}

static SECItem *op5lic_reader_get_content(struct op5lic_reader_t *lr, struct op5lic_error_t *err) {
	SECItem *item;
	NSSCMSMessage *cms_message;
	if (!lr->fp) {
		return NULL;
	}

	if ((cms_message = op5lic_reader_decode_license(lr)) == NULL) {
		op5lic_error_set(err, OP5LIC_STATUS_DECODING_ERROR, lr->filename);
		return NULL;
	}

	if (!op5lic_reader_license_is_signed(cms_message)) {
		op5lic_error_set(err, OP5LIC_STATUS_UNSIGNED, lr->filename);
		return NULL;
	}

	if (!op5lic_reader_license_signature_is_valid(cms_message, lr->trusted_certificates, lr->ntrusted)) {
		op5lic_error_set(err, OP5LIC_STATUS_INVALID_SIGNATURE, lr->filename);
		return NULL;
	}

	item = NSS_CMSMessage_GetContent(cms_message);
	if (item && item->data && item->len) {
		return item;
	}
	log_err("BUG: Failed to get content from succesfully decoded CMS message");
	op5lic_error_set(err, OP5LIC_STATUS_DECODING_ERROR, lr->filename);
	return NULL;
}

static LicenseStatus luastate_finalize(lua_State *L, struct op5lic_error_t *err) {
	int status;
	/* Call the script, to put all methods in the global scope */
	status = lua_pcall(L, 0, 0, 0);
	if (status) {
		/* If something went wrong, Lua error message is at the top of */
		/* the stack */
		log_err("Failed to run script: %s",
				lua_tostring(L, -1));

		op5lic_error_set(err, op5lic_error_from_luastatus(status), lua_tostring(L, -1));
		return op5lic_error_from_luastatus(status);
	}

	return OP5LIC_STATUS_OK;
}

LicenseStatus op5lic_reader_load_v1_v2_fallback(const char *license_path,
		lua_State *L, struct op5lic_error_t *err) {
	extern const char *g_license_backbox_lua_buffer;
	int ret = 0;
	LicenseStatus status;
	ezxml_t doc;

	/*Read legacy license*/
	doc = ezxml_parse_file(license_path);
	if (doc != NULL) {

		status = op5lic_backbox_license_verify(doc, err);
		if (status != OP5LIC_STATUS_OK) {
			ezxml_free(doc);
			op5lic_error_set(err, status, license_path);
			return err->status;
		}

		/*Transform elements to Lua table entries and
		 * load into a global table in lua_State L */
		op5lic_backbox_load_doc(L, doc);
		ezxml_free(doc);
	}
	else {
		log_warn("Initialising implicit Trial license, since the license file (%s) was not found", license_path);
		op5lic_backbox_init_implicit_trial(L);
	}
	ret = luaL_loadbuffer(L, g_license_backbox_lua_buffer, strlen(g_license_backbox_lua_buffer), "license content");
	if (ret != 0) {
		op5lic_error_set(err, op5lic_error_from_luastatus(ret), "v1/v2 backwards compatibility layer");
		return err->status;
	}
	return luastate_finalize(L, err);
}

LicenseStatus op5lic_reader_load_lua(struct op5lic_reader_t *lr,
		lua_State *L, struct op5lic_error_t *err) {
	int ret = 0;
	SECItem *item;
	item = op5lic_reader_get_content(lr, err);
	if (!item) {
		log_err("Failed to load license content (%s)", op5lic_strerror(err));
		return err->status;
	}

	xfree(lr->lua_buffer);

	lr->lua_buffer = xcalloc(item->len, 1);
	memcpy(lr->lua_buffer, item->data, item->len);
	ret = luaL_loadbuffer(L, lr->lua_buffer, item->len, "license content");
	if (ret != 0) {
		op5lic_error_set(err, op5lic_error_from_luastatus(ret), lr->filename);
		return err->status;
	}

	return luastate_finalize(L, err);
}
